#define _WIN32_WINDOWS 0x0444
#include <windows.h>
#include <time.h>
#include "zmouse.h"
#include "czerwony_kapturek_jbwheel_WheelTool.h"

#include <map>
using namespace std;

typedef map<HWND, DWORD> Window2Proc;

typedef DWORD (CALLBACK* WPROC)(HWND h, UINT u, LPARAM l, WPARAM w);

Window2Proc windows;

//handle of this dll module
HANDLE hm = NULL;
//handle to debugging console
HANDLE console = NULL;
//temporary handle to console
HANDLE tmpConsole = NULL;
//hook procedure for sent
HHOOK hhSend = NULL;
//ID of JBuilder process
DWORD myProcess = 0;

//message for Win95
UINT uMSH_MOUSEWHEEL = 0;

//jbWheel opentool object instance
jobject jo;
//JNI enviroment pointer
JNIEnv * env = NULL;
//pointer to Java VM
JavaVM* vm = NULL;
//id of the java method to call
jmethodID mid;

//print debugging messages
void trace( char* msg );

//main dll function
BOOL APIENTRY
DllMain( HANDLE hModule,
         DWORD  ul_reason_for_call,
         LPVOID /*lpReserved*/
        )
{
    hm = hModule;
    switch (ul_reason_for_call)
    {
        case DLL_PROCESS_ATTACH:
        case DLL_THREAD_ATTACH:
        case DLL_THREAD_DETACH:
            break;

        case DLL_PROCESS_DETACH:
            if (tmpConsole)
            {
                FreeConsole();
            }
            if (hhSend)
            {
                ::UnhookWindowsHookEx( hhSend );
                hhSend = 0;
            }
            break;
    }
    return TRUE;
}

//process messagess intercepted by hooks
void _stdcall
processMessage(bool w95, long lParam, long wParam)
{
    trace(w95 ? "MSH_MOUSEWHEEL" : "WM_MOUSEWHEEL" );
    if (!env)
    {
        vm->AttachCurrentThread( (void**)&env, NULL );
    //}
        // I'm not sure if storing method ID is valid
        jclass jc = env->GetObjectClass( jo );
        mid = env->GetMethodID( jc, "onWheelMoved", "(IIII)V");
    }

    short delta;
    short keys = 0;
    if (!w95)
    {
        delta = (short)HIWORD(wParam);
        keys = (short)LOWORD(wParam);
    }
    else
    {
        if ( GetKeyState( VK_SHIFT )&0xFF00 )
            keys |= MK_SHIFT;
        if ( GetKeyState( VK_CONTROL )&0xFF00 )
            keys |= MK_CONTROL;
        delta = (short)wParam;
    }

    env->CallVoidMethod( jo, mid, delta,
                (short)LOWORD(lParam),
                (short)HIWORD(lParam),
                keys /*+ (alt ? 0x8000 : 0)*/ );
}


LRESULT CALLBACK
mouseWndProc(
    HWND hwnd,        // handle to window
    UINT uMsg,        // message identifier
    WPARAM wParam,    // first message parameter
    LPARAM lParam)    // second message parameter
{
    switch (uMsg)
    {
        case WM_MOUSEWHEEL:
        {
            processMessage(false, lParam, wParam);
        }
        return 0;

        case WM_DESTROY:
            trace("WM_DESTROY !!!!!!!!!!!! this can't happen");
        default:
        {
            if (uMsg == uMSH_MOUSEWHEEL)
            {
                processMessage(true,lParam, wParam);
                return 0;
            }

            //trace("wndproc: default");
            Window2Proc::iterator it;
            if ((it = windows.find(hwnd)) != windows.end())
            {
                DWORD ret = ((WPROC)(*it).second)(hwnd, uMsg, wParam, lParam);
                return ret;
            }
            else
            {
                trace("Window not in the map !!!!!!!!!!!!! can't happen");
                return DefWindowProc(hwnd, uMsg, wParam, lParam);
            }
        }
    }
}


void
onCreate(HWND hwnd)
{
    DWORD process;
    GetWindowThreadProcessId(hwnd, &process);
    if (process == myProcess)
    {
        Window2Proc::iterator it;
        if ((it = windows.find(hwnd)) == windows.end())
        {
            long proc = GetWindowLongPtr(hwnd, GWLP_WNDPROC);
            char b[255];
            wsprintf(b, "registering %x, proc %x", hwnd, proc);
            trace(b);
            if (!IsBadCodePtr((int (__stdcall *)())proc))
            {
                windows[hwnd] = proc;
                SetWindowLong(hwnd, GWLP_WNDPROC, (long)mouseWndProc);
                char buff[255];
                wsprintf(buff, "registered, %i windows in the map", windows.size());
                trace(buff);
            }
            else
            {
                trace("invalid proc!!!");
            }
        }
        else
        {
            trace("already registered");
        }
    }
    else
    {
        trace("--------------NOT MY PROCESS");
    }
}


void
onDestroy(HWND hwnd)
{
    DWORD process;
    GetWindowThreadProcessId(hwnd, &process );
    if (process == myProcess)
    {
        Window2Proc::iterator it;
        if ((it = windows.find(hwnd)) != windows.end())
        {
            trace( "unregistering" );
            SetWindowLongPtr(hwnd, GWLP_WNDPROC, (*it).second);
            windows.erase(hwnd);
            char buff[255];
            wsprintf(buff, "unregistered, %i windows left", windows.size());
            trace(buff);
        }
        else
        {
            trace("not registered");
        }
    }
}


//hook for messages sent to the window
long _stdcall
sendHookProc(int code, unsigned int wParam, long lParam)
{
    long ret = CallNextHookEx(hhSend, code, wParam, lParam);

    if (code >= 0)
    {
        CWPSTRUCT* pMsg = (CWPSTRUCT*)lParam;
        if (pMsg->message==WM_CREATE)
        {
            trace("WM_CREATE");
            onCreate(pMsg->hwnd);
        }
        if (pMsg->message==WM_DESTROY)
        {
            trace("WM_DESTROY");
            onDestroy(pMsg->hwnd);
        }
    }

    return ret;
}


void
clearWindowsMap()
{
    Window2Proc::iterator it = windows.begin();
    while (it != windows.end())
    {
        SetWindowLongPtr((*it).first, GWLP_WNDPROC, (*it).second);
        it++;
    }
    windows.clear();
}


/*
 * Class:     czerwony_kapturek_jbwheel_WheelTool
 * Method:    startCaptureWheelMessages
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_startCaptureWheelMessages(JNIEnv * env1, jobject jo1)
{
    uMSH_MOUSEWHEEL = RegisterWindowMessage(MSH_MOUSEWHEEL);

    jo = env1->NewGlobalRef( jo1 );
    env1->GetJavaVM( &vm );

    hhSend = ::SetWindowsHookEx( WH_CALLWNDPROC, sendHookProc, (HINSTANCE)hm, NULL );
    if (!hhSend)
    {
        char    buff[255];
        wsprintf( buff, "Failed to register message hook for sent messagess (%08X)", GetLastError() );
        trace( buff );
    }

    myProcess = ::GetCurrentProcessId();

    clearWindowsMap();
}


/*
 * Class:     czerwony_kapturek_jbwheel_WheelTool
 * Method:    stopCaptureWheelMessages
 * Signature: ()V
 */
JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_stopCaptureWheelMessages(JNIEnv *, jobject)
{
    trace("stopCaptureWheelMessages");
    env->DeleteGlobalRef( jo );
    jo = NULL;
    env = NULL;
    vm = NULL;
    mid = 0;
    if (hhSend)
    {
        ::UnhookWindowsHookEx( hhSend );
        hhSend = 0;
    }

    clearWindowsMap();
}


JNIEXPORT void JNICALL
Java_czerwony_kapturek_jbwheel_WheelTool_setDebug(JNIEnv*, jobject, jboolean debug )
{
    trace("setDebug()");
    console = NULL;
    if (debug)
    {
        if (!tmpConsole)
        {
            AllocConsole();
            tmpConsole = GetStdHandle( STD_ERROR_HANDLE );
        }
        console = tmpConsole;
    }
}


//print debug messages to the console
void
trace( char* msg )
{
    if ( console )
    {
        printf("jbWheel.dll: ");
        printf( msg );
        printf("\n");
    }
}



